import reprlib
reprlib.aRepr.maxtuple = 999
reprlib.aRepr.maxlist = 999
reprlib.aRepr.maxarray = 999
reprlib.aRepr.maxdict = 999
reprlib.aRepr.maxset = 999
reprlib.aRepr.maxfrozenset = 999
reprlib.aRepr.maxdeque = 999
reprlib.aRepr.maxstring = 999
reprlib.aRepr.maxlong = 999
reprlib.aRepr.maxother = 999
